/* 	===========================================================================
P_Ch3 Phase 4: Timer_MMSS
Control Unit that controls the 
Counter_MMSS with CE, UD_L and multiplexed 7-segment output and the possibility 
to select the CLK source that runs the counter:
CLK_Flag  --> 1 Hz 
CLK_MUX_Flag --> 20 Hz
 
TARGET MICROCONTROLLER: ATMEGA8535 
(but the example is fully compatible with many other Atmel processors)
 
NOTE: This code is not optimised, but in some way extended in order to clarify 
and make easy the understanding of the basic operations. At this level of 
complexity, having to interconnect several individual "components", it is 
preferable to organise the code using several source files, functions and 
global and local variables.  
  	======================================================================== */

#include <avr/io.h>
//#include <stdlib.h>    
#include <avr/interrupt.h>

// Let's see where all the ATmega8535 microcontroller declarations are located.

// C:\Program Files (x86)\Atmel\Atmel Studio 6.2\devices\ATmega8535.xml
/* #include "Atmel Toolchain/AVR8 GCC/Native/3.4.1056/avr8-gnu-toolchain
							/avr/include/avr/iom8535.h"
*/
/* 	===========================================================================
	Definitions 
   	======================================================================== */
 //------------------ Inputs definitions ----------------------
// Example hardware:  Counter_MMSS.pdsprj
// 
// The port PORTA where the input "ST_Up" is located
#define TS_Up_PORT	PINA 	
// The pin number (PA0) where the ST_Up is located is the mask
#define TS_Up_pin		0b00000001		

// The port PORTA where the input "ST_Down" is located
#define TS_Down_PORT	PINA
// The pin number (PA1) where the ST_Up is located is the mask
#define TS_Down_pin		0b00000010

// The port PORTA where the input "Sel_CLK" is located
//#define SelCLK_PORT	PINA
// The pin number (PA2) where the CE is located is the mask
//#define SelCLK_pin 	0b00000100

// The CLK input, can be defined the same way, however, it won't be read but 
// instead, the interface of this input will be conducted through 
// the INT0 interrupt (edge detection, programmable rising/falling edge)

// The PORTC where the input CLK is located
// #define CLK_PORT PIND		
// The pin number where the input CLK is located (PD2/INT0)
// #define CLK_pin	0b00000100	

// ------------------- Outputs definitions ----------------------
// The port where 7-segment outputs of the counter are connected,
// and so it will be written
#define PORT_7SEG PORTC		

// Other definitions related to the counter system:
// The list of ASCII codes to be used to save counter states 
#define 	Num_0 		'A'		//States enumeration. "Char"
#define 	Num_1 		'B'
#define 	Num_2 		'C'
#define 	Num_3 		'D'
#define 	Num_4 		'E'
#define 	Num_5 		'F'
#define 	Num_6 		'G'
#define 	Num_7 		'H'
#define 	Num_8 		'I'
#define 	Num_9 		'J'

// States for the control unit 
#define	Idle 			'M'
#define	Set_Time_Up		'N'
#define	Set_Time_Down 	'O'
#define	Timing 			'P'
#define	Alarm_Set 		'Q'

/* 	===========================================================================
	Function prototypes
   	======================================================================== */
//#  define ISR(vector, [attributes]) 
// Already defined in the header <avr/interrupt.h>
// General functions (top level): 
void init_system (void);
void read_inputs(void);
void write_outputs(void);
char BCD_to_7seg (char); // takes the BCD number as a parameter

// Counter seconds block/component: 
char state_logic_DSU(void);
char output_logic_DSU(void);

char state_logic_DST(void);
char output_logic_DST(void);

// Counter minutes block/component:
char state_logic_DMU(void);
char output_logic_DMU(void);

char state_logic_DMT(void);
char output_logic_DMT(void);


// Basically the Counter_MMSS
void datapath(void); 

// the finite state machine to control the datapath and to generate  output signals 
char state_logic_CU(void);
char output_logic_CU(void);

/* ============================================================================
	Global variables (RAM memory) to make the system compatible and as 
	hardware-independent as possible. 
	For easy watching/debugging and processing in C 
  	======================================================================== */
// General top level variables in the schematic:
char Value_CEM;		// [0 0 0 0 0 0 0 CEM] wire to chain seconds and minutes 
char Value_TCH;		// Final terminal count 1 hour

char Value_CE;  //General count enable
char Value_UD_L;  //General Up-Down
char Value_SelCLK;	// [0 0 0 0 0 0 0 SelCLK] -- Select the CLK source

// CLK generation signals
char Value_Freq_div3;  // Variables for the CLK_generator in the ISR()
char Value_Freq_div20;
char CLK_Flag;	//The flag that will be set from the interrupt service routine
// to run the counters
char CLK_MUX_Flag;//Flag that will be set from the interrupt service routine
// to be used for multiplexing the digits

char Value_CA; // to decoder the digits 1110, 1101, 1011, 0111, 1110, etc.

// ------------- Seconds counter block ----------------------------------------------
char Value_SS_CET;		// [0 0 0 0 0 0 0 CET]
char Value_SS_TC10;		// [0 0 0 0 0 0 0 TC10] 
char Value_SS_TC6;		// [0 0 0 0 0 0 0 TC6]
char Value_SS_TC60;		// [0 0 0 0 0 0 0 TC60]
	
char Value_DSU;	// [0 0 0 0 Q3 Q2 Q1 Q0] to store the counter number in BCD
char Value_DST;
	// The state variable of the FSM which will be updated ("asynchronously")
	// every time that "state_logic_DSU()" is executed
char current_state_DSU;		
char current_state_DST;	

// Decimal points -- not used in this application.
char Value_DST_DP = 0; // [0 0 0 0 0 0 0 DP], but not used in this design
char Value_DSU_DP = 0;

char Value_DSU_7seg; // Variable to store the units in 7-segment code
char Value_DST_7seg; // Variable to store the tens in 7-segment code


// ------------- Minutes counter block ----------------------------------------------
char Value_MM_CET;		// [0 0 0 0 0 0 0 CET]
char Value_MM_TC10;		// [0 0 0 0 0 0 0 TC10]
char Value_MM_TC6;		// [0 0 0 0 0 0 0 TC6]
char Value_MM_TC60;		// [0 0 0 0 0 0 0 TC60]

char Value_DMU;	// [0 0 0 0 Q3 Q2 Q1 Q0] to store the counter number in BCD
char Value_DMT;
// The state variable of the FSM which will be updated ("asynchronously")
// every time that "state_logic_DSU()" is executed
char current_state_DMU;
char current_state_DMT;

char Value_DMT_DP = 0;
char Value_DMU_DP = 0;	// Used for representing the intermittent seconds LED

char Value_DMU_7seg; // Variable to store the units in 7-segment code
char Value_DMT_7seg; // Variable to store the tens in 7-segment code

// The current state of the control unit and the new variables related to the 
// inputs that has to be read and the ones that generates interrupts
char current_state_CU;
char Value_TS_Up;	
char Value_TS_Down;
char Start_Flag;
char Stop_Flag;

char Value_Timing_LED;
char Value_Alarm_LED;
char Value_S_LED;	// This is the intermittent 1 Hz signal 
char Value_Soundwave; 

// The last improvement may be to consider a CLK to move to synchronise 
// the control unit 
char CU_Flag;


/* 	===========================================================================
	Atmega8535 Configuration bit Fuses (datasheet page 238)
  	=========================================================================== 
 This definitions are copied from the "iom8535.h" header file:
	   #define FUSE_MEMORY_SIZE 2
  // Low Fuse Byte 
	   //CKSEL(3..0) = "1110" External resonator or crystal up to 8 MHz
	   #define FUSE_CKSEL0      (unsigned char)~_BV(0)	
	   #define FUSE_CKSEL1      (unsigned char)~_BV(1)
	   #define FUSE_CKSEL2      (unsigned char)~_BV(2)
	   #define FUSE_CKSEL3      (unsigned char)~_BV(3)
	   //SUT(1..0) = "11" Crystal Oscillator, slowly rising power
	   #define FUSE_SUT0        (unsigned char)~_BV(4)	
	   #define FUSE_SUT1        (unsigned char)~_BV(5)
	   // Brown out detector enable
	   #define FUSE_BODEN       (unsigned char)~_BV(6)	
	   #define FUSE_BODLEVEL    (unsigned char)~_BV(7)// BODLEVEL = "0" ==> 4V
// Brown-out Reset. The MCU is reset when the supply voltage VCC is below the
// Brown-out Reset threshold (VBOT) and the Brown-out Detector is enabled.
	   
#define LFUSE_DEFAULT (FUSE_CKSEL1 & FUSE_CKSEL2 
		& FUSE_CKSEL3 & FUSE_SUT0 & FUSE_SUT1)
			// "0001" "00" Internal 1 MHz calibrated oscillator, 
	   // High Fuse Byte 
	   #define FUSE_BOOTRST     (unsigned char)~_BV(0)	//Select Reset Vector
	   //The default value of BOOTSZ1..0 results in maximum Boot Size
	   #define FUSE_BOOTSZ0     (unsigned char)~_BV(1)	
	   #define FUSE_BOOTSZ1     (unsigned char)~_BV(2)	
	   //EEPROM memory is preserved through the Chip Erase
	   #define FUSE_EESAVE      (unsigned char)~_BV(3)	
	   // When CKOPT is programmed, the Oscillator output will oscillate will 
	   // a full rail-to-rail swing on the output.
	   #define FUSE_CKOPT       (unsigned char)~_BV(4)	
	   // Enable Serial Programming
	   #define FUSE_SPIEN       (unsigned char)~_BV(5)	
	   // Enable the Watch Dog timer
	   #define FUSE_WDTON       (unsigned char)~_BV(6)	
	   	// Compatible to the AT90S8535
	   #define FUSE_S8535C      (unsigned char)~_BV(7)
	   #define HFUSE_DEFAULT (FUSE_BOOTSZ0 & FUSE_BOOTSZ1 & FUSE_SPIEN)
*/

// ============================================================================
// So, let's program the configuration bits for a given 
// developing board and AVR chip. Do not change these values
    FUSES =
    {
		// let's use an external oscillator "1110", "01", and BODEN:
			    //.low = (FUSE_BODEN & FUSE_BODLEVEL),
		.low = (FUSE_CKSEL0 & FUSE_SUT1 & FUSE_BODEN & FUSE_BODLEVEL),
		.high = (FUSE_BOOTSZ0 & FUSE_BOOTSZ1 & FUSE_SPIEN & FUSE_EESAVE),
    }; 

/* 	===========================================================================
	Main function 
  	======================================================================== */
int main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
		read_inputs ();
		
		if (CU_Flag == 1){	// To synchronise the CU
			state_logic_CU();
			output_logic_CU();				
			CU_Flag = 0;
		}
	
		write_outputs(); // Write the ports	
		datapath();
		
    }
}	
/* 	===========================================================================
	Function definitions
  	======================================================================== */

//*****************************************************************************
//Initialisation operations (setup) 
//*****************************************************************************

void init_system (void){
// Hardware configuration bits for polling/writing and interrupts

// Application program starts here:
	PORTA 	= 0x00;		// Let's not use the Pull-up resistor in the inputs 
						// (we'll wire external resistors like R4)
	DDRA 	= 0b11111100; // All outputs except PA1 (ST_Up), PA0 (ST_Down) 
	
	PORTD 	= 0x00;		// Let's not use the Pull-up resistor in PD2
						// Our CLK input will include R_CLK pull-up resistor
	// All outputs except the PD2 pin which is our CLK (interrupt INT0)
	// and the PD3 pin which is our Start (interrupt INT1)
	
	// PD7, PD6, PD5 and PD4 are the CA_L3, CA_L2, CA_L1 and CA_L0
	DDRD 	= 0b11110011;
	
	PORTD = (PIND & 0b00001111 ) | 0b11110000 ;// All OFF	
								
	PORTC 	= 0x00;
	DDRC	= 0xFF; // All outputs 

	PORTB 	= 0x00;		// Let's not use the Pull-up resistor in the inputs
	DDRB 	= 0b11111011; // All outputs except PB2: input INT2 (Stop)

		// Configuring the external interrupts: 
	// Trigger INT0 on falling edge (interrupt after clicking the pushbutton)
	MCUCR = 1<<ISC01 | 0<<ISC00;	

	// Trigger INT1 on falling edge (interrupt after clicking the pushbutton)
	MCUCR = 1<<ISC11 | 0<<ISC10;	
	
	// Trigger INT2 on falling edge (interrupt after clicking the pushbutton)
	MCUCSR = 0<<ISC2 ;

// Enable ONLY the external interrupt CLK and Start 
// Because clicking at this point the Stop, doesn't have to affect. 
	GICR = 0<<INT2 | 1<<INT1 | 1<<INT0;		
	
	sei(); // Global interrupts allowed
	
	current_state_DSU = Num_0;	//Reset all the counters MMSS
	current_state_DST = Num_0;	
	current_state_DMU = Num_0;
	current_state_DMT = Num_0;
	
    CLK_Flag = 0;		// The flag to run the counters 
    CLK_MUX_Flag = 0;	// The flag to run the 7-segment digit multiplexer
						
	Value_SS_TC10 = 0;		// Set the initial output values
	Value_DSU = 0;	// Init the counter values
	Value_DST  = 0;	
	Value_DMU = 0;
	Value_DMT  = 0;
	Value_CA = 0; 
	
	current_state_CU = Idle; // Initialise the Control Unit
    Start_Flag = 0;		// The flag to run the counters
    Stop_Flag = 0;	// The flag to run the 7-segment digit multiplexer
	
	write_outputs();
}


//*****************************************************************************
//Interrupt Service Routine to attend the external INT0 (The PD2 pin)
//*****************************************************************************
ISR(INT0_vect){
	cli();		// clear all sources of interrupt while attending this one
	
	// To generate a soundwave of 60 Hz; 
	Value_Soundwave = Value_Soundwave ^ 0b00000001; //toggle 
	CU_Flag = 1;	
	
	Value_Freq_div3 ++;	// If external interrupt, increment the counter_mod3 
	if (Value_Freq_div3 == 3){
		CLK_MUX_Flag = 1;	// If  counter_mod3 overflows, set the flag
				
		Value_Freq_div3 = 0;	// reset the counter_mod3
		
		Value_Freq_div20 ++;	// and increment the counter_mod20
		
		if (Value_Freq_div20 == 20){
			CLK_Flag = 1;		// If counter_mod20 overflows, set the flag
			Value_Freq_div20 = 0;	// reset the counter_mod20		
		}			
	}

// Select the CLK for the counter 
	if (Value_SelCLK == 1 ){
		CLK_Flag = CLK_MUX_Flag; 
	}
	

//Enable general interrupts, so that the system can be interrupted again	                     
	sei();	
}


//*****************************************************************************
//Interrupt Service Routine to attend the external INT1 (The PD3 pin) 
// Start button
//*****************************************************************************
ISR(INT1_vect){
	cli();		
		Start_Flag = 1;	// If  counter_mod3 overflows, set the flag
//Enable general interrupts, so that the system can be interrupted again	                     
	sei();	
}

//*****************************************************************************
//Interrupt Service Routine to attend the external INT2 (The PB2 pin) 
// Stop button
//*****************************************************************************
ISR(INT2_vect){
	cli();		
		Stop_Flag = 1;	// If  counter_mod3 overflows, set the flag
//Enable general interrupts, so that the system can be interrupted again	                     
	sei();	
}

	
//*****************************************************************************
// Read (poll) the input pins (level sensitive) 
//*****************************************************************************
void read_inputs (void){
	__asm ("nop");	// do nothing by now

//  Polling --> Read the ports and set the "char" variable
// Let's read the CE input pin and set the convenient variable "Value_CE" 
// using the masking technique: 
	if ( TS_Up_PORT & TS_Up_pin )	
		Value_TS_Up = 0;
	else
		Value_TS_Up = 1;
//*****************************************************************************
// Let's read the UD_L input pin and set the convenient variable "Value_UD_L"
// using the masking technique:
	if ( TS_Down_PORT & TS_Down_pin )
		Value_TS_Down = 0;
	else
		Value_TS_Down = 1;

}

//*****************************************************************************
// Convert from BCD to seven segments (active high). A simple truth table
//*****************************************************************************
char BCD_to_7seg (char Value_BCD_in){
	//char Value_error = 0;
	char Value_7seg = 0;	// [0 a b c d e f g] to store each segment

	switch (Value_BCD_in) {
		case 0:
			//				abcdefg
			//              -------
			Value_7seg = 0b01111110;
		break;
		case 1:
			Value_7seg = 0b00110000;
		break;
		case 2:
			Value_7seg = 0b01101101;
		break;
		case 3:
			Value_7seg = 0b01111001;
		break;
		case 4:
			Value_7seg = 0b00110011;
		break;
		case 5:
			Value_7seg = 0b01011011;
		break;
		case 6:
			Value_7seg = 0b01011111;
		break;
		case 7:
			Value_7seg = 0b01110000;
		break;
		case 8:
			Value_7seg = 0b01111111;
		break;
		case 9:
			Value_7seg = 0b01110011;
		break;
	}	
	return (Value_7seg);
};
		
//*****************************************************************************
// Write the output variables to the corresponding pins and ports 
//*****************************************************************************
void write_outputs(void){
	char Value_out; // A convenient local buffer

	Value_out = Value_Soundwave & Value_Alarm_LED; 
	PORTA = Value_Alarm_LED << 4 | Value_Timing_LED << 3 | Value_out <<5  ;
	
	//PORTA = ~(Value_TCH << 3); // Write the TCH

	// ---------------- Multiplexing 7-segment digits -------------------------
		// Write the CA_L(1..0) and the corresponding 7-segment digit
		// alternatively under the CLK_MUX_Flag rate
	if (CLK_MUX_Flag == 1){

		switch (Value_CA) {
			case 0:  // The DSU
				Value_out = Value_DSU_7seg | (Value_DSU_DP <<7);
				PORT_7SEG = ~Value_out;
				PORTD = 0b11100000 ;// CA_L0 bit (PD4) to '0'
			break;
			case 1:	// The DST
				Value_out = Value_DST_7seg | (Value_DST_DP <<7);
				PORT_7SEG = ~Value_out;
				PORTD = 0b11010000 ;// CA_L1 bit (PD5) to '0'
			break;
			
			case 2:	// The DMU
					// Seconds are intermittent only when timing
				Value_DMU_DP = Value_DMU_DP & Value_Timing_LED; 
				Value_out = Value_DMU_7seg | (Value_DMU_DP <<7);
				PORT_7SEG = ~Value_out;
				PORTD = 0b10110000 ;// CA_L2 bit (PD6) to '0'			
			break;
			
			case 3:	// The DMT
				Value_out = Value_DMT_7seg | (Value_DMT_DP <<7) ;
				PORT_7SEG = ~Value_out;
				PORTD = 0b01110000 ;// CA_L3 bit (PD7) to '0'			
			break;
		}

		Value_CA ++; // Counter_mod4 to decode digits
		if (Value_CA == 4) {
			Value_CA = 0;
		}
		
	CLK_MUX_Flag = 0;	
	}
};


//*****************************************************************************
//SECONDS FSM - UNITS - Output logic routine 
//*****************************************************************************
char output_logic_DSU(void){
	unsigned char Value_error = 0;
// Calculate the outputs
	switch (current_state_DSU) { // Set the outputs:a truth table
		case Num_0:
			Value_DSU = 0;
		break;
		case Num_1:
			Value_DSU = 1;
		break;
		case Num_2:
			Value_DSU = 2;
		break;
		case Num_3:
			Value_DSU = 3;
		break;
		case Num_4:
			Value_DSU = 4;
		break;
		case Num_5:
			Value_DSU = 5;
		break;
		case Num_6:
			Value_DSU = 6;
		break;
		case Num_7:
			Value_DSU = 7;
		break;
		case Num_8:
			Value_DSU = 8;
		break;
		case Num_9:
			Value_DSU = 9;
		break;
		
		default:
			Value_error = 1;	// If no states have been selected, 
								// there is an Value_error
								// and the systems must be stopped
		break;
	}
	// Lets set the value of the Value_SS_TC10:
	if ( (current_state_DSU == Num_9 && Value_UD_L == 1 && Value_CE == 1 ) || 
		 (current_state_DSU == Num_0 && Value_UD_L == 0 && Value_CE == 1 ) ) 
		Value_SS_TC10 = 1;
	else
		Value_SS_TC10 = 0;

	return (Value_error);
}


//*****************************************************************************
//SECONDS FSM - UNITS - State variable logic routine 
//*****************************************************************************
char state_logic_DSU(void){
	char Value_error = 0;
// A behavioural way to interpret and translate the truth table of the CC1 
	//read_inputs (); // Lets poll all the variables which have to be sampled  
					// (level sensitive)
	// Do something only when there is an CLK's active edge detection
	if (Value_CE == 1){	// The system has to count			
		if (Value_UD_L == 1){	// Count UP
			switch (current_state_DSU) {
				case Num_0:
				current_state_DSU = Num_1;
				break;
				case Num_1:
				current_state_DSU = Num_2;
				break;
				case Num_2:
				current_state_DSU = Num_3;
				break;
				case Num_3:
				current_state_DSU = Num_4;
				break;
				case Num_4:
				current_state_DSU = Num_5;
				break;
				case Num_5:
				current_state_DSU = Num_6;
				break;
				case Num_6:
				current_state_DSU = Num_7;
				break;
				case Num_7:
				current_state_DSU = Num_8;
				break;
				case Num_8:
				current_state_DSU = Num_9;
				break;
				case Num_9:
				current_state_DSU = Num_0;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		else{	// Down 
			switch (current_state_DSU) {
				case Num_0:
				current_state_DSU = Num_9;
				break;
				case Num_1:
				current_state_DSU = Num_0;
				break;
				case Num_2:
				current_state_DSU = Num_1;
				break;
				case Num_3:
				current_state_DSU = Num_2;
				break;
				case Num_4:
				current_state_DSU = Num_3;
				break;
				case Num_5:
				current_state_DSU = Num_4;
				break;
				case Num_6:
				current_state_DSU = Num_5;
				break;
				case Num_7:
				current_state_DSU = Num_6;
				break;
				case Num_8:
				current_state_DSU = Num_7;
				break;
				case Num_9:
				current_state_DSU = Num_8;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
	}
	return (Value_error);
}
	
		
//*****************************************************************************
//SECONDS FSM - TENS Output logic routine 
//*****************************************************************************
char output_logic_DST(void){
	unsigned char Value_error = 0;
	
// Calculate the outputs
	
	switch (current_state_DST) {	// Set the outputs for every state:a truth table
		case Num_0:
			Value_DST = 0;
		break;
		case Num_1:
			Value_DST = 1;
		break;
		case Num_2:
			Value_DST = 2;
		break;
		case Num_3:
			Value_DST = 3;
		break;
		case Num_4:
			Value_DST = 4;
		break;
		case Num_5:
			Value_DST = 5;
		break;
		
		default:
			Value_error = 1;	// If no states have been selected, there is an Value_error
		// and the systems must be stopped
		break;
	}
	
	// Lets set the value of the Value_SS_TC10:
	if ( (current_state_DST == Num_5 && Value_UD_L == 1 && Value_SS_CET == 1 ) || 
		 (current_state_DST == Num_0 && Value_UD_L == 0 && Value_SS_CET == 1 ) ) 
		Value_SS_TC6 = 1;
	else
		Value_SS_TC6 = 0;
			
	return (Value_error);
}		

//*****************************************************************************
//SECONDS FSM - TENS State variable logic routine 
//*****************************************************************************
char state_logic_DST(void){
	char Value_error = 0;
// A behavioural way to interpret and translate the truth table of the CC1 
	//read_inputs (); // Lets poll all the variables which have to be sampled  
					// (level sensitive)
	// Do something only when there is an CLK's active edge detection
	if (Value_SS_CET == 1){	// The system has to count			
		if (Value_UD_L == 1){	// Count UP
			switch (current_state_DST) {
				case Num_0:
				current_state_DST = Num_1;
				break;
				case Num_1:
				current_state_DST = Num_2;
				break;
				case Num_2:
				current_state_DST = Num_3;
				break;
				case Num_3:
				current_state_DST = Num_4;
				break;
				case Num_4:
				current_state_DST = Num_5;
				break;
				case Num_5:
				current_state_DST = Num_0;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		else{	// Down 
			switch (current_state_DST) {
				case Num_0:
				current_state_DST = Num_5;
				break;
				case Num_1:
				current_state_DST = Num_0;
				break;
				case Num_2:
				current_state_DST = Num_1;
				break;
				case Num_3:
				current_state_DST = Num_2;
				break;
				case Num_4:
				current_state_DST = Num_3;
				break;
				case Num_5:
				current_state_DST = Num_4;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		
	}
	return (Value_error);
}

//*****************************************************************************
//MINUTES FSM - UNITS - State variable logic routine 
//*****************************************************************************
char state_logic_DMU(void){
	char Value_error = 0;
// A behavioural way to interpret and translate the truth table of the CC1 
	//read_inputs (); // Lets poll all the variables which have to be sampled  
					// (level sensitive)
	// Do something only when there is an CLK's active edge detection
	if (Value_CEM == 1){	// The system has to count			
		if (Value_UD_L == 1){	// Count UP
			switch (current_state_DMU) {
				case Num_0:
				current_state_DMU = Num_1;
				break;
				case Num_1:
				current_state_DMU = Num_2;
				break;
				case Num_2:
				current_state_DMU = Num_3;
				break;
				case Num_3:
				current_state_DMU = Num_4;
				break;
				case Num_4:
				current_state_DMU = Num_5;
				break;
				case Num_5:
				current_state_DMU = Num_6;
				break;
				case Num_6:
				current_state_DMU = Num_7;
				break;
				case Num_7:
				current_state_DMU = Num_8;
				break;
				case Num_8:
				current_state_DMU = Num_9;
				break;
				case Num_9:
				current_state_DMU = Num_0;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		else{	// Down 
			switch (current_state_DMU) {
				case Num_0:
				current_state_DMU = Num_9;
				break;
				case Num_1:
				current_state_DMU = Num_0;
				break;
				case Num_2:
				current_state_DMU = Num_1;
				break;
				case Num_3:
				current_state_DMU = Num_2;
				break;
				case Num_4:
				current_state_DMU = Num_3;
				break;
				case Num_5:
				current_state_DMU = Num_4;
				break;
				case Num_6:
				current_state_DMU = Num_5;
				break;
				case Num_7:
				current_state_DMU = Num_6;
				break;
				case Num_8:
				current_state_DMU = Num_7;
				break;
				case Num_9:
				current_state_DMU = Num_8;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
	}
	return (Value_error);
}
	
//*****************************************************************************
//MINUTES FSM - UNITS - Output logic routine 
//*****************************************************************************
char output_logic_DMU(void){
	unsigned char Value_error = 0;
// Calculate the outputs
	switch (current_state_DMU) {	// Set the outputs for every state:a truth table
		case Num_0:
			Value_DMU = 0;
		break;
		case Num_1:
			Value_DMU = 1;
		break;
		case Num_2:
			Value_DMU = 2;
		break;
		case Num_3:
			Value_DMU = 3;
		break;
		case Num_4:
			Value_DMU = 4;
		break;
		case Num_5:
			Value_DMU = 5;
		break;
		case Num_6:
			Value_DMU = 6;
		break;
		case Num_7:
			Value_DMU = 7;
		break;
		case Num_8:
			Value_DMU = 8;
		break;
		case Num_9:
			Value_DMU = 9;
		break;
		
		default:
			Value_error = 1;	// If no states have been selected, there is an Value_error
		// and the systems must be stopped
		break;
	}
	// Lets set the value of the Value_SS_TC10:
	if ( (current_state_DMU == Num_9 && Value_UD_L == 1 && Value_CEM == 1 ) || 
		 (current_state_DMU == Num_0 && Value_UD_L == 0 && Value_CEM == 1 ) ) 
		Value_MM_TC10 = 1;
	else
		Value_MM_TC10 = 0;

	return (Value_error);
}

//*****************************************************************************
//MINUTES FSM - TENS - State variable logic routine 
//*****************************************************************************
char state_logic_DMT(void){
	char Value_error = 0;
// A behavioural way to interpret and translate the truth table of the CC1 
	//read_inputs (); // Lets poll all the variables which have to be sampled  
					// (level sensitive)
	// Do something only when there is an CLK's active edge detection
	if (Value_MM_CET == 1){	// The system has to count			
		if (Value_UD_L == 1){	// Count UP
			switch (current_state_DMT) {
				case Num_0:
				current_state_DMT = Num_1;
				break;
				case Num_1:
				current_state_DMT = Num_2;
				break;
				case Num_2:
				current_state_DMT = Num_3;
				break;
				case Num_3:
				current_state_DMT = Num_4;
				break;
				case Num_4:
				current_state_DMT = Num_5;
				break;
				case Num_5:
				current_state_DMT = Num_0;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		else{	// Down 
			switch (current_state_DMT) {
				case Num_0:
				current_state_DMT = Num_5;
				break;
				case Num_1:
				current_state_DMT = Num_0;
				break;
				case Num_2:
				current_state_DMT = Num_1;
				break;
				case Num_3:
				current_state_DMT = Num_2;
				break;
				case Num_4:
				current_state_DMT = Num_3;
				break;
				case Num_5:
				current_state_DMT = Num_4;
				break;

				default:
				Value_error = 1;
				// If no states have been selected, there is an Value_error
				// and the systems must be stopped
				break;
			}
		}
		
	}
	return (Value_error);
}


//*****************************************************************************
//MINUTES FSM - TENS - Output logic routine 
//*****************************************************************************
char output_logic_DMT(void){
	unsigned char Value_error = 0;
	
// Calculate the outputs
	
	switch (current_state_DMT) {// Set outputs for every state:a truth table
		case Num_0:
			Value_DMT = 0;
		break;
		case Num_1:
			Value_DMT = 1;
		break;
		case Num_2:
			Value_DMT = 2;
		break;
		case Num_3:
			Value_DMT = 3;
		break;
		case Num_4:
			Value_DMT = 4;
		break;
		case Num_5:
			Value_DMT = 5;
		break;
		
		default:
			Value_error = 1;	// If no states have been selected, there is an Value_error
		// and the systems must be stopped
		break;
	}
	
	// Lets set the value of the Value_SS_TC10:
	if ( (current_state_DMT == Num_5 && Value_UD_L == 1 && Value_MM_CET == 1 ) || 
		 (current_state_DMT == Num_0 && Value_UD_L == 0 && Value_MM_CET == 1 ) ) 
		Value_MM_TC6 = 1;
	else
		Value_MM_TC6 = 0;
			
	return (Value_error);
}		


/* 	===========================================================================
	Counter_MMS which is controlled by the control_unit() 
  	======================================================================== */

void datapath (void){  
	
	if (CLK_Flag == 1){
	
		// ---------------- Seconds -----------------------------------
		output_logic_DSU();  // SS units counter
		state_logic_DSU();
				
		if (Value_DSU <10) {// Convert to 7 segments
			Value_DSU_7seg = BCD_to_7seg (Value_DSU);
		}
				
		Value_SS_CET= Value_SS_TC10; // The wire that chains the counters
				
		output_logic_DST();	// SS tens counter
		state_logic_DST();
				
		if (Value_DST <6) {// Convert to 7 segments
			Value_DST_7seg = BCD_to_7seg (Value_DST);
		}
				
		// Lets set the value of the Value_SS_TC60, so that later
		// we can enhance this chain of counters
		if ( Value_SS_TC10 == 1 && Value_SS_TC6 == 1)
		Value_SS_TC60 = 1;
		else
		Value_SS_TC60 = 0;
		// ------------------------------------------------------------
				
		Value_CEM = Value_SS_TC60; // The wire that chains the counters
				
		// ---------------- Minutes -----------------------------------
		output_logic_DMU();  // MM units counter
		state_logic_DMU();
				
		if (Value_DMU <10) {// Convert to 7 segments
			Value_DMU_7seg = BCD_to_7seg (Value_DMU);
		}
				
		Value_MM_CET= Value_MM_TC10; // The wire that chains the counters
				
		output_logic_DMT();	// MM tens counter
		state_logic_DMT();
				
		if (Value_DMT <6) {// Convert to 7 segments
			Value_DMT_7seg = BCD_to_7seg (Value_DMT);
		}
				
		// Lets set the value of the Value_SS_TC60, so that later
		// we can enhance this chain of counters
		if ( Value_MM_TC10 == 1 && Value_MM_TC6 == 1)
		Value_MM_TC60 = 1;
		else
		Value_MM_TC60 = 0;
		// ------------------------------------------------------------
				
		// Lets set the value of the Value_TCH, so that later
		// we can enhance this chain of counters
		if ( Value_SS_TC60 == 1 && Value_MM_TC60 == 1)
		Value_TCH = 1;
		else
		Value_TCH = 0;
				
		// Seconds LED intermittent generation
		Value_DMU_DP = ~Value_DMU_DP;
				
			
		// Very important --> Clear the CLK_flag,
		// so it can be set again in the next falling edge in PD2
		CLK_Flag = 0;
	}	
	
}


/* 	===========================================================================
	Control_unit() state logic 
  	======================================================================== */

char state_logic_CU(void){
	char error = 0;
	switch (current_state_CU) {
		case Idle:
			if (Start_Flag == 1 ){
				current_state_CU = Timing;					
				Start_Flag = 0; // Clear the Flag
			}
			else if (Value_TS_Up == 1 )
				current_state_CU = Set_Time_Up;
			else if (Value_TS_Down == 1 )
				current_state_CU = Set_Time_Down;
			else 
				current_state_CU = Idle;						
		break;
		
		case Set_Time_Up:
			if (Value_TS_Up  == 1 )
				current_state_CU = Set_Time_Up;
			else
				current_state_CU = Idle;
		break;
		
		case Set_Time_Down:
			if (Value_TS_Down  == 1 )
				current_state_CU = Set_Time_Down;
			else
				current_state_CU = Idle;
		break;

		case Timing:
			if (Value_TCH == 1 ){
				current_state_CU = Alarm_Set;
					// Enable the 3 external interrupt sources:
					// So that it can be stopped when a Click is detected.
				GICR = 1<<INT2 | 1<<INT1 | 1<<INT0;				
			}
			else
				current_state_CU = Timing;
		break;
		
		case Alarm_Set:
			if (Stop_Flag == 1 ){
				current_state_CU = Idle;
				Stop_Flag = 0; // Clear the Flag  
				// Enable ONLY the external interrupt CLK and Start
				// Because clicking at this point the Stop, doesn't have to affect.
				GICR = 0<<INT2 | 1<<INT1 | 1<<INT0;				
			}
			else
				current_state_CU = Alarm_Set;
		break;

		default:
			error = 1;	// If no states have been selected, there is an error
						// and the systems must be stopped
		break;
	}
	return (error);
	
}


/* 	===========================================================================
	Control_unit() output logic
  	======================================================================== */
char output_logic_CU(void){
	char error = 0;
	
// Calculate the outputs
	switch (current_state_CU) {		// Set the outputs for every state:a truth table
		case Idle:
			// Control signals for the datapath
			Value_CE = 0;
			Value_UD_L = 0; // Don't care
			Value_SelCLK = 0; // Don't care
			// Output signals
			Value_Timing_LED = 0 ;
			Value_Alarm_LED = 0 ; // Which also generates the sound wave
			
		break;
	
		case Set_Time_Up:
			Value_CE = 1;  // Let it count
			Value_UD_L = 1; // Counting Up at higher speed
			Value_SelCLK = 1; // Select the 20 Hz CLK
			Value_Timing_LED = 0 ;
			Value_Alarm_LED = 0 ; 
		break;
		
		case Set_Time_Down:
			Value_CE = 1;  // Let it count
			Value_UD_L = 0; // Counting Down at higher speed
			Value_SelCLK = 1; // Select the 20 Hz CLK
			Value_Timing_LED = 0 ;
			Value_Alarm_LED = 0 ;
		break;
		
		case Timing:
			Value_CE = 1;  // Let it count
			Value_UD_L = 0; // Counting Down at 1 Hz 
			Value_SelCLK = 0; // Select the 1 Hz CLK
			Value_Timing_LED = 1 ; // Indicating this process of timing
								   // Let it represent the 1 Hz intermittently 
			Value_Alarm_LED = 0 ;
		break;
		
		case Alarm_Set:
	
			Value_CE = 0;  // Stop the counter 
			Value_UD_L = 0; // Don't care  
			Value_SelCLK = 0; // Don't care 
			Value_Timing_LED = 0 ; // End of timing period 
			Value_Alarm_LED = 1 ; // LED and sound wave 
			current_state_DSU = Num_0;	//Reset all the counters MMSS
			current_state_DST = Num_0;
			current_state_DMU = Num_0;
			current_state_DMT = Num_0;
			
		break;
	

		default:
			error = 1;		// If no states have been selected, there is an error
		// and the systems must be stopped
		break;
	}
	
	return (error);
}
	
